/*
*  * Copyright (C) 2001-2004 Alejandro Aguilar Sierra <asierra@servidor.unam.mx>
*  *  and Dan Dennedy <dan@dennedy.org>
*  *
*  * This program is free software; you can redistribute it and/or modify
*  * it under the terms of the GNU General Public License as published by
*  * the Free Software Foundation; either version 2 of the License, or
*  * (at your option) any later version.
*  *
*  * This program is distributed in the hope that it will be useful,
*  * but WITHOUT ANY WARRANTY; without even the implied warranty of
*  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  * GNU General Public License for more details.
*  *
*  * You should have received a copy of the GNU General Public License
*  * along with this program; if not, write to the Free Software Foundation,
*  * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*  */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <iostream>
#include <gnome.h>

extern "C"
{
#include "interface.h"
#include "support.h"
	GtkWidget *dvtitler_window = NULL;
}

#include "dvtitler.h"

#include "textblock.h"

DVTitler::DVTitler()
{
	dvtitler_window = window = create_wintitler();
	text = 0;
	titler = 0;
	pixbuf = 0;
	pih = pfh = piv = pfv = 0;
	interlaced = true;
	GnomeColorPicker *colorFG = GNOME_COLOR_PICKER( my_lookup_widget( GTK_WIDGET( window ), "colorpickerfg" ) );
	GnomeColorPicker *colorBG = GNOME_COLOR_PICKER( my_lookup_widget( GTK_WIDGET( window ), "colorpickerbg" ) );
	gnome_color_picker_set_i8( colorFG, 0xff, 0xff, 0xff, 0xff );
	gnome_color_picker_set_i8( colorBG, 0, 0, 0, 0 );
}

DVTitler::~DVTitler()
{
	if ( pixbuf != 0 )
		g_object_unref( pixbuf );
	gtk_widget_destroy( window );
	delete titler;
}


char *DVTitler::GetDescription( ) const
{
	return "DV Titler " VERSION;
}


void DVTitler::AttachWidgets( GtkBin *bin )
{
	gtk_widget_reparent( ( GTK_BIN( window ) ) ->child, GTK_WIDGET( bin ) );
}

void DVTitler::DetachWidgets( GtkBin *bin )
{
	gtk_widget_reparent( ( GTK_BIN( bin ) ) ->child, GTK_WIDGET( window ) );
}

void DVTitler::InterpretWidgets( GtkBin *bin )
{
	GtkMenu * ihMenu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( my_lookup_widget( GTK_WIDGET( window ), "optionmenuih" ) ) ) );
	GtkMenu *ivMenu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( my_lookup_widget( GTK_WIDGET( window ), "optionmenuiv" ) ) ) );
	GtkMenu *fhMenu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( my_lookup_widget( GTK_WIDGET( window ), "optionmenufh" ) ) ) );
	GtkMenu *fvMenu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( my_lookup_widget( GTK_WIDGET( window ), "optionmenufv" ) ) ) );

	GtkTextView *titleEntry = GTK_TEXT_VIEW( my_lookup_widget( GTK_WIDGET( window ), "text" ) );
	GnomeColorPicker *colorFG = GNOME_COLOR_PICKER( my_lookup_widget( GTK_WIDGET( window ), "colorpickerfg" ) );
	GnomeColorPicker *colorBG = GNOME_COLOR_PICKER(
	                                my_lookup_widget( GTK_WIDGET( window ), "colorpickerbg" ) );


	// Initial horizontal position
	GtkWidget *active_item = gtk_menu_get_active( ihMenu );
	ih = g_list_index ( GTK_MENU_SHELL ( ihMenu ) ->children, active_item ) ;

	// Initial vertical position
	active_item = gtk_menu_get_active( ivMenu );
	iv = g_list_index ( GTK_MENU_SHELL ( ivMenu ) ->children, active_item ) ;

	// Final horizontal position
	active_item = gtk_menu_get_active( fhMenu );
	fh = g_list_index ( GTK_MENU_SHELL ( fhMenu ) ->children, active_item ) ;

	// Final vertical position
	active_item = gtk_menu_get_active( fvMenu );
	fv = g_list_index ( GTK_MENU_SHELL ( fvMenu ) ->children, active_item ) ;

	if ( fh > 4 ) // no change
		fh = ih;

	if ( fv > 4 ) // no change
		fv = iv;

	pad = gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON( my_lookup_widget( GTK_WIDGET( window ), "spinpad" ) ) );
	int space = gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON( my_lookup_widget( GTK_WIDGET( window ), "spinspace" ) ) );

	GtkMenu *alignMenu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( my_lookup_widget( GTK_WIDGET( window ), "alignmenu" ) ) ) );
	active_item = gtk_menu_get_active( alignMenu );
	int align = g_list_index ( GTK_MENU_SHELL ( alignMenu ) ->children, active_item ) ;

	GnomeFontPicker *fontpicker = GNOME_FONT_PICKER( my_lookup_widget( GTK_WIDGET( window ), "fontpicker" ) );
	const char *fontname = gnome_font_picker_get_font_name( fontpicker );

	xoff = gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON( my_lookup_widget( GTK_WIDGET( window ), "spinxoff" ) ) );
	yoff = gtk_spin_button_get_value_as_int( GTK_SPIN_BUTTON( my_lookup_widget( GTK_WIDGET( window ), "spinyoff" ) ) );

	// Obtain the text
	g_free( text );
	GtkTextIter start;
	GtkTextIter end;
	GtkTextBuffer* buffer = gtk_text_view_get_buffer( titleEntry );
	gtk_text_buffer_get_iter_at_offset( buffer, &start, 0 );
	gtk_text_buffer_get_iter_at_offset( buffer, &end, -1 );
	text = gtk_text_buffer_get_text( buffer, &start, &end, FALSE );

	interlaced = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( my_lookup_widget( GTK_WIDGET( window ), "checkbutton_interlaced" ) ) );

	delete titler;
	titler = new TextBlock();
	titler->setFont( fontname );
	titler->setPad( pad );
	titler->setSpace( space );
	titler->setAlign( align );
	titler->setText( text );

	// Obtain the color
	gnome_color_picker_get_i8( colorFG, &colorfg.r, &colorfg.g, &colorfg.b, &colorfg.a );
	gnome_color_picker_get_i8( colorBG, &colorbg.r, &colorbg.g, &colorbg.b, &colorbg.a );
}

void DVTitler::FilterFrame( uint8_t *io, int width, int height, double position, double frame_delta )
{
	int x, y;

	if ( position == 0 )
	{

		if ( pixbuf != 0 )
			g_object_unref( pixbuf );

		pixbuf = titler->getPixbuf( colorfg, colorbg );
		if ( pixbuf == 0 )
		{
			GtkWidget * dialog = gtk_message_dialog_new(
		                         GTK_WINDOW( NULL ),
		                         GTK_DIALOG_DESTROY_WITH_PARENT,
		                         GTK_MESSAGE_ERROR,
		                         GTK_BUTTONS_CLOSE,
		                         _("Rendering failed. Are you using bad markup?") );
			gtk_dialog_run( GTK_DIALOG( dialog ) );
			gtk_widget_destroy( dialog );
			return;
		}

		// scale to preview size
		if ( width < 720 )
		{
			GdkPixbuf * temp = pixbuf;
			GdkPixbuf *scaled = gdk_pixbuf_scale_simple( pixbuf, gdk_pixbuf_get_width( pixbuf ) / 4,
			                    gdk_pixbuf_get_height( pixbuf ) / 4, GDK_INTERP_HYPER );
			pixbuf = scaled;
			g_object_unref( temp );
			pad /= 4;
			xoff /= 4;
			yoff /= 4;
			space /= 4;
		}
		// scale to adjust for NTSC sample aspect ratio
		else if ( height < 576 )
		{
			GdkPixbuf * temp = pixbuf;
			GdkPixbuf *scaled = gdk_pixbuf_scale_simple( pixbuf,
				(gint) ( (float) gdk_pixbuf_get_width( pixbuf ) * 720.0/640.0 ),
				gdk_pixbuf_get_height( pixbuf ), GDK_INTERP_HYPER );
			pixbuf = scaled;
			g_object_unref( temp );
		}
		// scale to adjust for PAL sample aspect ratio
		else
		{
			GdkPixbuf * temp = pixbuf;
			GdkPixbuf *scaled = gdk_pixbuf_scale_simple( pixbuf,
				(gint) ( (float)gdk_pixbuf_get_width( pixbuf ) * 720.0/768.0 ),
				gdk_pixbuf_get_height( pixbuf ), GDK_INTERP_HYPER );
			pixbuf = scaled;
			g_object_unref( temp );
		}
		
		int w = gdk_pixbuf_get_width ( pixbuf );
		int h = gdk_pixbuf_get_height ( pixbuf );

		// if Left Off -title width, else if Right Off frame width,
		// else (Left = 0, Center = 1, Right = 2) * difference / 2
		// pih = initial horizontal position
		// pfh = final horizontal position
		pih = ( ih == 3 ) ? -w : ( ih > 3 ) ? width : ( float ) ih * ( width - w ) / 2.0;
		pfh = ( fh == 3 ) ? -w : ( fh > 3 ) ? width : ( float ) fh * ( width - w ) / 2.0;

		// if Top Off -title height, else if Bottom Off frame height, else
		// else (Top = 0, Middle = 1, Bottom = 2) * difference / 2
		// piv = initial vertical position
		// pfv = final vertical position
		piv = ( iv == 3 ) ? -h : ( iv > 2 ) ? height : ( float ) iv * ( height - h ) / 2.0;
		pfv = ( fv == 3 ) ? -h : ( fv > 2 ) ? height : ( float ) fv * ( height - h ) / 2.0;

		// store the width and height for future use in other methods
		frame_w = width;
		frame_h = height;
	}

	if ( pixbuf )
	{
		for ( int field = 0; field < ( interlaced ? 2 : 1 ); ++ field )
		{
			// Offset the position based on which field we're looking at ...
			const double field_position = position + ( 1 - field ) * frame_delta * 0.5;
				
			// apply the progress factor, position, to the path as an offset to the
			// initial position plus apply user-supplied offsets
			x = ( int ) ( field_position * ( pfh - pih ) + pih ) + xoff;
			y = ( int ) ( field_position * ( pfv - piv ) + piv ) + yoff;

			//std::cerr << "drawing title on field " << field << " at " << x << ", " << y << std::endl;
			drawPixbuf( io, // destination buffer
					x, y,
					width * 3, // destination buffer stride
					( 1 - field ) );
		}
	}
}


/** Draw the pixbuf from the titler into the frame RGB buffer

    \param io Pointer to RGB buffer
    \param x  Horizontal position
    \param y  Vertical position
    \param frame_stride The number of bytes in a single row of the frame RGB buffer
    \param field 0 for the upper field or progressive frame, 1 for the lower field
*/
void DVTitler::drawPixbuf( uint8_t *io, int x, int y, int frame_stride, int field )
{
	if ( 3 * x > frame_stride )
		return ;

	int xb = 0, yb = 0, title_w, title_h, title_stride;

	title_w = gdk_pixbuf_get_width ( pixbuf );
	title_h = gdk_pixbuf_get_height ( pixbuf );
	title_stride = gdk_pixbuf_get_rowstride( pixbuf );

	// optimization point - no work to do
	if ( ( x < 0 && -x >= title_w ) || ( y < 0 && -y >= title_h ) )
		return;

	// crop overlay off the left edge of frame
	if ( x < 0 )
	{
		xb = -x;
		title_w -= xb;
		x = 0;
	}
	// crop overlay beyond right edge of frame
	else if ( x + title_w > frame_w )
		title_w = frame_w - x;

	// crop overlay off the top edge of the frame
	if ( y < 0 )
	{
		yb = -y;
		title_h -= yb;
	}
	// crop overlay below bottom edge of frame
	else if ( y + title_h > frame_h )
		title_h = frame_h - y;

	// initialise pointer into overlay buffer based on cropping
	uint8_t *pixels = gdk_pixbuf_get_pixels( pixbuf ) + xb * 4 + yb * title_stride;

	// henceforth, bring y into the frame space
    y = y < 0 ? 0 : y;
	
	// initialise pointer into frame buffer
	uint8_t *base = io + ( x < 0 ? 0 : x ) * 3 + y * frame_stride;

	// special care is taken to make sure frame buffer offset is aligned to the correct field.
	// field 0 = lower field and y should be odd (y is 0-based).
	// field 1 = upper field and y should be even.
	if ( interlaced && ( ( field == 0 && y % 2 == 0 ) || ( field == 1 && y % 2 != 0 ) ) )
		base += frame_stride;

	// now do the compositing only to cropped extents
	for ( int j = 0; j < title_h; j += ( interlaced ? 2 : 1 ) )
	{
		uint8_t *p = base + ( j * frame_stride );
		uint8_t *q = pixels + ( j * title_stride );
		for ( int i = 0; i < title_w; i++ )
		{
			float a = ( float ) *( q + 3 ) / 255.0;
			*p = ( int ) ( a * ( *q++ ) + ( 1 - a ) * ( *p ) );
			p++;
			*p = ( int ) ( a * ( *q++ ) + ( 1 - a ) * ( *p ) );
			p++;
			*p = ( int ) ( a * ( *q++ ) + ( 1 - a ) * ( *p ) );
			p++;
			q++;
		}
	}
}


void DVTitler::drawRectangle( uint8_t *io, int x, int y, int w, int h, int frame_stride )
{
	uint8_t * base = io + frame_stride * y + x * 4;

	for ( int j = 0; j < h; j++ )
	{
		uint8_t *p = base;
		for ( int i = 0; i < w; i++ )
		{
			*p ++ = colorbg.r;
			*p ++ = colorbg.g;
			*p ++ = colorbg.b;
			*p ++ = colorbg.a;
		}
		base += frame_stride;
	}
}


extern "C"
{

	GDKImageFilter *GetImageFilter( int index )
	{

		switch ( index )
		{
		case 0:
			return new DVTitler();
		}

		return NULL;
	}
}

